/*
 * ExpandedBitBaseOutputStream.java
 *
 * Created on 5.1 2004, 20:48
 */

package saiko.compression;

import java.io.ByteArrayOutputStream;
import java.io.FilterOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.math.BigInteger;

/**
 * Class to provide expansion of bit base of data stream
 *
 * @author  Dusan Saiko
 * @see saiko.compression.ExpandedBitBaseInputStream
 */
public class ExpandedBitBaseOutputStream extends FilterOutputStream {
    
    /**
     * the header sign on the compressed data
     */
    public final static String header="DSCEBB";
    
    private ByteArrayOutputStream buffer;  //data buffer 
    private int  bufferSize; //data buffer size
    private byte[] prev_alphabet; // previously used alphabet
    
    /** 
     * Constructor of new instance of ExpandedBitBaseOutputStream
     * Constructor sets the default buffer size to be 1 KB.
     *
     * @param aOutput (OutputStream) - enclosed output stream
     * @exception  IOException if an I/O error occurs
    */
    public ExpandedBitBaseOutputStream(OutputStream aOutput) throws IOException
    {
        this(aOutput, 1024);
    }

    /** 
     * Constructor of new instance of ExpandedBitBaseOutputStream
     *
     * @param aOutput (OutputStream) - enclosed output stream
     * @param aBufferSize (int)     -  size of the buffer, maximum 65535 bytes
     * @exception  IOException if an I/O error occurs
    */
    public ExpandedBitBaseOutputStream(OutputStream aOutput, int aBufferSize) throws IOException
    {
        super(aOutput);
        out.write(header.getBytes());
        bufferSize=Math.min(Math.max(1024,aBufferSize),65536);
        buffer=new ByteArrayOutputStream(bufferSize);
    }

    /**
     * Flush the data to the output
     */
    private synchronized void flushBuffer() throws IOException {
        // if buffer is empty, do nothing
        if(buffer==null || buffer.size()==0) return;
        
        //get the data
        byte data[]=buffer.toByteArray();
        
        //get the stats of order 0 and sort it (will be sorted 'alphabetically')
        Statistics[] stats=new SourceStatistics(data).getStats(0,true);
        
        //reindex data according to the stats
        for(int i=0; i<data.length; i++) {
            for(int n=0; n<stats.length; n++) {
                if(stats[n].data[0]==data[i]) {
                    data[i]=(byte)n;
                    break;
                }
            }
        }
        
        //compute the number the data represents
        BigInteger base=new BigInteger(String.valueOf(stats.length));
        
        //start with zero
        BigInteger value=new BigInteger("0");
        for(int i=0; i<data.length; i++) {
            //multiply the previous result with the base size
            value=value.multiply(base);
            //ad new lowest value to the big number
            value=value.add(new BigInteger(String.valueOf(((int)(char)data[i])&0xFF)));
        }
        
        //write out the alphabet
        byte[] alphabet=new byte[stats.length];
        for(int i=0; i<alphabet.length; i++) alphabet[i]=stats[i].data[0];
        
        //compare the alphabet with previous one
        boolean eq=false;
        if(prev_alphabet!=null && prev_alphabet.length==alphabet.length) {
            eq=true;
            for(int i=0; i<alphabet.length; i++) 
                if(alphabet[i]!=prev_alphabet[i]) { 
                    eq=false; break; 
                }
        }
        
        if(eq) {
            //if previous alphabet should be used
            out.write(0);            
        } else {
            //if alphabet differs from the previous one
            out.write(alphabet.length);
            out.write(alphabet);
            prev_alphabet=alphabet;
        }

        //write out the length of the original buffer;
        int length=data.length;
        for(int i=0; i<2; i++) {
            int v=(length>>(8*(1-i)))&0xFF;
            out.write(v);
        }

        //write out the length of the number data;
        data=value.toByteArray();
        length=data.length;
        for(int i=0; i<2; i++) {
            out.write((length>>(8*(1-i)))&0xFF);
        }
        
        //write out the number
        out.write(data);
        
        //reset buffer
        buffer.reset();
    }

    /**
     * Writes the specified byte to this compressed output stream. 
     *
     * @param      aValue (int)  the byte to be written.
     * @exception  IOException  if an I/O error occurs.
     */
    public synchronized void write(int aValue) throws IOException {
        buffer.write(aValue);
        if(buffer.size()>=bufferSize) {
            flushBuffer();
        }
    }

  
    /**
     * Flushes this buffered output stream. This forces any buffered 
     * output bytes to be written out to the underlying output stream. 
     *
     * @exception  IOException  if an I/O error occurs.
     * @see        java.io.FilterOutputStream#out
     */
    public synchronized void flush() throws IOException {
        flushBuffer();
    } 
    

}
