/*
 * HuffmanOutputStream.java
 *
 * Created on 21.1 2004, 20:48
 */

package saiko.compression;

import java.io.ByteArrayOutputStream;
import java.io.FilterOutputStream;
import java.io.IOException;
import java.io.OutputStream;

/**
 * Class to provide huffman coding compression
 *
 * @author  Dusan Saiko
 * @see saiko.compression.HuffmanInputStream
 */
public class HuffmanOutputStream extends FilterOutputStream {
    

    
    private ByteArrayOutputStream buffer;  //data buffer 
    private int  bufferSize; //data buffer size
    
    /** 
     * Constructor of new instance of ExpandedBitBaseOutputStream
     * Constructor sets the default buffer size to be 1 KB.
     *
     * @param aOutput (OutputStream) - enclosed output stream
     * @exception  IOException if an I/O error occurs
    */
    public HuffmanOutputStream(OutputStream aOutput) throws IOException
    {
        this(aOutput, 256*256);
    }

    /** 
     * Constructor of new instance of ExpandedBitBaseOutputStream
     *
     * @param aOutput (OutputStream) - enclosed output stream
     * @param aBufferSize (int)     -  size of the buffer, maximum 65535 bytes
     * @exception  IOException if an I/O error occurs
    */
    public HuffmanOutputStream(OutputStream aOutput, int aBufferSize) throws IOException
    {
        super(aOutput);
        bufferSize=Math.min(Math.max(1024,aBufferSize),256*256);
        buffer=new ByteArrayOutputStream(bufferSize);
    }

    /**
     * Flush the data to the output
     */
    private synchronized void flushBuffer() throws IOException {
        // if buffer is empty, do nothing
        if(buffer==null || buffer.size()==0) return;
        
        byte data[] = buffer.toByteArray();
        
        //get statistic, do not sort them
        Statistics stats[]=new SourceStatistics(data).getStats(1,false);

        //create and fill the huffman binary tree
        HuffmanBinaryTree tree=new HuffmanBinaryTree();
        if(stats!=null) { //stats are null for 0 length file
            for(int i=0; i<stats.length; i++) {
                tree.addElement(stats[i].data, stats[i].count);
            }
            tree.compute();
        }
        
        //output the original buffer size
        int length=data.length-1;
        for(int i=0; i<2; i++) {
            int v=(length>>(8*(1-i)))&0xFF;
            out.write(v);
        }
        
        //output the statistics */
        out.write(stats.length-1);
        
        for(int i=0; i<stats.length; i++) {
            out.write(stats[i].data[0]);
            length=stats[i].count-1;
            for(int n=0; n<2; n++) {
                int v=(length>>(8*(1-n)))&0xFF;
                out.write(v);
            }
        }
        
        //transform the data
        String currentData="";
        int i=0;
        while(i<data.length) {
            String code=tree.getCode(new byte[]{data[i]});
            currentData+=code;
            if(currentData.length()>7) {
                String currentByte = currentData.substring(0,8);
                currentData=currentData.substring(8);
                int value=Integer.parseInt(currentByte,2);
                out.write(value);
            }
            i++;
        }
        //write out what rest in the last coded byte
        while(currentData.length()>0) {
                length=Math.min(currentData.length(),8);
                String currentByte = currentData.substring(0,length);
                while(currentByte.length()<8) currentByte+="0";
                currentData=currentData.substring(length);
                int value=Integer.parseInt(currentByte,2);
                out.write(value);
        }
        
        //reset buffer
        buffer.reset();
    }

    /**
     * Writes the specified byte to this compressed output stream. 
     *
     * @param      aValue (int)  the byte to be written.
     * @exception  IOException  if an I/O error occurs.
     */
    public synchronized void write(int aValue) throws IOException {
        buffer.write(aValue);
        if(buffer.size()>=bufferSize) {
            flushBuffer();
        }
    }

  
    /**
     * Flushes this buffered output stream. This forces any buffered 
     * output bytes to be written out to the underlying output stream. 
     *
     * @exception  IOException  if an I/O error occurs.
     * @see        java.io.FilterOutputStream#out
     */
    public synchronized void flush() throws IOException {
        flushBuffer();
    }     
}
