/*
 * PBEOutputStream.java
 *
 * Created on 27.12. 2003, 23:10
 */

package saiko.compression;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FilterOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.security.InvalidParameterException;
import java.security.NoSuchAlgorithmException;

import javax.crypto.Cipher;
import javax.crypto.CipherOutputStream;
import javax.crypto.SecretKey;
import javax.crypto.SecretKeyFactory;
import javax.crypto.spec.PBEKeySpec;
import javax.crypto.spec.PBEParameterSpec;


/**
 * Class to provide Password Based Encryption output stream
 *
 * @author  Dusan Saiko
 * @see javax.crypto.Cipher
 * @see javax.crypto.CipherInputStream
 */
public class PBEOutputStream extends FilterOutputStream {
    
    /**
     * the header sign on the crypto data stream
     * DSPBEV1
     */
    public final static String header="DSPBE1";
    
    /**
     * number of iterations to use to create the salt for PBEParameterSpec
     */
    public final static int ITERATIONS=20;
    
    
    /** 
     * Constructor of new instance of BPEOutputStream
     *
     * @param aOutput (OutputStream) - output stream to write write the result into
     * @param aPassword (String) - password, minimum 9 characters
    */
    public PBEOutputStream(OutputStream aOutput, String aPassword) 
    throws  IOException,
            InvalidParameterException,
            NoSuchAlgorithmException,
            java.security.spec.InvalidKeySpecException,
            java.security.InvalidKeyException,
            javax.crypto.NoSuchPaddingException,
            java.security.InvalidAlgorithmParameterException
    {
        super(aOutput);
        
        if(aPassword==null || aPassword.length()<9) throw 
            new InvalidParameterException("Password has to be at least 9 chars");

        // Begin by creating a salt of 64 bits (8 byte)
        byte [] salt = aPassword.substring(0,8).getBytes();

        // Create the PBEKeySpec with the given password          
        PBEKeySpec keySpec = new PBEKeySpec(aPassword.toCharArray());

        // Create a SecretKeyFactory for PBEWithSHAAndTwofish
        SecretKeyFactory keyFactory =
            SecretKeyFactory.getInstance("PBEWithMD5AndDES");

        // Create our key
        SecretKey key = keyFactory.generateSecret (keySpec); 

        // Now create a parameter spec for our salt and iterations
        PBEParameterSpec paramSpec = new PBEParameterSpec (salt, ITERATIONS);

        //Create a cipher and initialize it for encryption
        Cipher cipher = Cipher.getInstance("PBEWithMD5AndDES");
        cipher.init (Cipher.ENCRYPT_MODE, key, paramSpec);

        out=new CipherOutputStream(aOutput, cipher);
        out.write(header.getBytes());
    }

    public static void main(String aArg[]) throws Exception {
        InputStream input=new FileInputStream("c:/file.txt");
        OutputStream output= new PBEOutputStream(new FileOutputStream("c:/file.txt.cr"), "PASSWORD1");
        int i;
        while((i=input.read())>=0) output.write(i);
        input.close();
        output.flush();
        output.close();
    }
}
